//  -----------------------------------------------------------------------------
//
//  Copyright (C) 2012 Fons Adriaensen <fons@linuxaudio.org>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  -----------------------------------------------------------------------------


#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <getopt.h>
#include <stdio.h>
#include <jack/jack.h>
#include "audiofile.h"


enum { HELP, CHAN, CAF, WAV, AMB, BIT16, BIT24, FLOAT, REC, TRI, LIPS };
enum { MAXINP = 64, BLOCK = 4096 };


static struct option options [] = 
{
    { "help",  0, 0, HELP  },
    { "chan",  1, 0, CHAN  },
    { "caf",   0, 0, CAF   },
    { "wav",   0, 0, WAV   },
    { "amb",   0, 0, AMB   },
    { "16bit", 0, 0, BIT16 },
    { "24bit", 0, 0, BIT24 },
    { "float", 0, 0, FLOAT },
    { "rec",   0, 0, REC   },
    { "tri",   0, 0, TRI   },
    { "lips",  0, 0, LIPS  },
    { 0, 0, 0, 0 }
};


static void help (void)
{
    fprintf (stderr, "\njack_export %s\n", VERSION);
    fprintf (stderr, "(C) 2012 Fons Adriaensen  <fons@linuxaudio.org>\n");
    fprintf (stderr, "Multitrack export from ardour2.\n");
    fprintf (stderr, "Recording starts when Jack enters freewheeling mode,\n");
    fprintf (stderr, "and stops when Jack returns to normal mode.\n");
    fprintf (stderr, "Usage: jack_export <options> <output file>.\n");
    fprintf (stderr, "Options:\n");
    fprintf (stderr, "  --help : Display this text.\n");
    fprintf (stderr, "  --chan  <number of channels>\n");
    fprintf (stderr, "  Output file type:      --wav, --amb, --caf\n");
    fprintf (stderr, "  Output sample format:  --16bit, --24bit, --32bit, --float\n");
    fprintf (stderr, "  Dither type (16 bit):  --rec, --tri, --lips\n");
    fprintf (stderr, "  The default output file format is WAV, 2 channels, 24-bit.\n");
    exit (1);
}


static int           type = Audiofile::TYPE_WAV;
static int           form = Audiofile::FORM_24BIT;
static int           dith = Audiofile::DITHER_NONE; 
static int           nchan = 2;
static int           fsamp = 0;
static int           fragm = 0;
static int           count = 0;
static int64_t       total = 0;
static bool          proc = false;
static bool          done = false;
static const char    *jname = 0;
static float         *out_buff = 0;
static Audiofile     *out_file = 0;
static jack_client_t *client = 0;
static jack_port_t   *inp_port [MAXINP];


static void procoptions (int ac, char *av [])
{
    int k;

    while ((k = getopt_long (ac, av, "", options, 0)) != -1)
    {
	switch (k)
	{
        case '?':
	case HELP:
	    help ();
	    break;
	case CHAN:
	    if (sscanf (optarg, "%d", &nchan) != 1)
	    {
		fprintf (stderr, "Illegal value for --nchan option: '%s'.\n", optarg);
		exit (1);
	    }
	    break;
	case WAV:
	    type = Audiofile::TYPE_WAV;
	    break;
	case AMB:
	    type = Audiofile::TYPE_AMB;
	    break;
	case CAF:
	    type = Audiofile::TYPE_CAF;
	    break;
	case BIT16:
	    form = Audiofile::FORM_16BIT;
	    break;
	case BIT24:
	    form = Audiofile::FORM_24BIT;
	    break;
	case FLOAT:
	    form = Audiofile::FORM_FLOAT;
	    break;
	case REC:
	    dith = Audiofile::DITHER_RECT;
	    break;
	case TRI:
	    dith = Audiofile::DITHER_TRIA;
	    break;
	case LIPS:
	    dith = Audiofile::DITHER_LIPS;
	    break;
 	}
    }
}


static void jack_shutdown (void *arg)
{
    proc = false;
    done = true;
    count = 0;
}


static void jack_freewheel (int state, void *arg)
{
    if (state)
    {
	proc = true;
    }
    else
    {
	proc = false;
	done = true;
    }
}


static int jack_process (jack_nframes_t nframes, void *arg)
{
    int    i, j;
    float  *p, *q;

    if (! proc) return 0;

    if (count + (int) nframes > BLOCK)
    {
	out_file->write (out_buff, count);
	count = 0;
    }

    for (i = 0; i < nchan; i++)
    {
	p = (float *)(jack_port_get_buffer (inp_port [i], nframes));
	q = out_buff + (count * nchan + i);
	for (j = 0; j < (int) nframes; j++)
	{
	    *q = p [j];
	    q += nchan;
	}
    }
    if (!total) printf ("Recording ...\n");
    count += (int) nframes;
    total += (int) nframes;

    return 0;
}


static int init_jack (void)
{
    jack_status_t       stat;
    int                 opts, i;
    char                s1 [64];
    char                s2 [64];

    opts = JackNoStartServer;
    if ((client = jack_client_open ("jack_export", (jack_options_t) opts, &stat)) == 0)
    {
        fprintf (stderr, "Can't connect to JACK\n");
        return 1;
    }

    jack_on_shutdown (client, jack_shutdown, 0);
    jack_set_freewheel_callback (client, jack_freewheel, 0);
    jack_set_process_callback (client, jack_process, 0);

    if (jack_activate (client))
    {
        fprintf(stderr, "Can't activate JACK.\n");
        return 1;
    }

    jname = jack_get_client_name (client);
    fsamp = jack_get_sample_rate (client);
    fragm = jack_get_buffer_size (client);

    for (i = 0; i < nchan; i++)
    {
	sprintf (s1, "in.%d", i + 1);
        inp_port [i] = jack_port_register (client, s1, JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	if (inp_port [i] == 0)
	{
	    fprintf (stderr, "Can't create input port '%s'.\n", s1);
	    return 1;
	}
        sprintf (s1, "ardour:export/out %d", i + 1);
        sprintf (s2, "jack_export:in.%d", i + 1);
	if (jack_connect (client, s1, s2))
	{
	    fprintf (stderr, "Warning: can't connect %s -> %s\n", s1, s2);
	}
    }

    return 0;
}


void cleanup (void)
{
    if (client)
    {
        jack_deactivate (client);
        jack_client_close (client);
    }
    delete[] out_buff;
    delete   out_file;
}


int main (int ac, char *av [])
{
    int i;

    for (i = 0; i < MAXINP; i++) inp_port [i] = 0;

    procoptions (ac, av);
    if (ac < optind + 1)
    {
        fprintf (stderr, "Missing output filename, try --help.\n");
	return 1;
    }
    if ((nchan < 1) || (nchan > MAXINP))
    {
        fprintf (stderr, "Illegal number of channels, should be 1...64.\n");
	return 1;
    }

    if (init_jack ())
    {
	return 1;
    }

    out_buff = new float [nchan * BLOCK];
    out_file = new Audiofile ();
    if (out_file->open_write (av [optind], type, form, fsamp, nchan))
    {
	fprintf (stderr, "Can't open output file '%s'.\n", av [optind]);
        cleanup ();
	return 1;
    }
    if (dith != Audiofile::DITHER_NONE) 
    {
	out_file->set_dither (dith);
    }

    puts ("Ready ...");
    while (! done) usleep (500000);
    if (count) out_file->write (out_buff, count);
    out_file->close ();
    printf ("Recorded %5.3lf seconds\n", (double) total / fsamp);
    puts ("Done");

    cleanup ();
    return 0;
}

